--- ./erts/emulator/beam/binary.c.old	2007-12-31 17:07:20.000000000 +0100
+++ ./erts/emulator/beam/binary.c	2008-02-04 21:31:28.000000000 +0100
@@ -29,6 +29,7 @@
 #include "big.h"
 #include "erl_binary.h"
 #include "erl_bits.h"
+#include <assert.h>
 
 Uint erts_allocated_binaries;
 erts_mtx_t erts_bin_alloc_mtx;
@@ -36,6 +37,7 @@
 static int list_to_bitstr_buf(Eterm obj, char* buf, int len);
 static Sint bitstr_list_len(Eterm obj);
 
+
 void
 erts_init_binary(void)
 {
@@ -666,3 +668,810 @@
     DESTROY_ESTACK(s);
     return (Sint) -1;
 }
+  
+/*
+ *  Boyer-Moore, the fastest known single pattern search algorithm(?)
+ *  The longer the needle, the faster the search. Movement of needle 
+ *  in haystack is decided as max of values from two precalculated 
+ *  tables, bad shift table and good suffix table
+ */
+
+BIF_RETTYPE single_pattern_search(byte* haystack, byte* needle, 
+                                  Uint startpos, Uint hlen, Uint nlen)
+{
+    int i, j, k;
+    int hpos; 
+
+   /* 
+    *  Create bad character shift table as per Boyer-Moore.
+    *  This is essentially a table saying how many steps we can take if no
+    *  match is found. It is set to -1 for characters which are not found
+    *  in the needle, which means that if examine a char in the haystack
+    *  and it is not in the needle we get to jump a full nlen-1-(-1) steps.
+    *  Worst case is if the haystack char is in the last position of the 
+    *  needle, which means we skip nlen-1-(i=(nlen-2)) = 1 chars.
+    */
+
+    int occ[UCHAR_MAX + 1]; 
+    for(i = 0; i < UCHAR_MAX + 1; i++)
+        occ[i] = -1;
+    for(i = 0; i < nlen - 1; ++i) 
+        occ[needle[i]] = i;
+
+    /*
+     *  For the second table, the "good suffix table", we first need to create
+     *  a list of suffixes in the needle. suffix[i] = the longest suffix of 
+     *  the needle up to i which is also a suffix of the whole needle
+     *  This table basically tells if a suffix of the needle reoccurs
+     *  within the needle, e.g. suffix abc reoccuring in dabcdeabc
+     */
+
+    //Variable size arrays may be gcc specific? Will other compilers be used?
+    int suffix[nlen], skip[nlen];
+
+    suffix[nlen-1] = nlen; /* The longest suffix of the needle is the needle */
+    j = 0;        /* j = end of latest longest suffix or i */
+    k = nlen - 1; /* k = start-1 of the latest longest suffix or i */
+    i = nlen - 2; /* loop counter for needle */
+    
+
+    while (i >= 0) {
+       /* If we are still *within* an old matched suffix and ... */
+       if (i > k && suffix[i + nlen - 1 - j] < i - k)
+          suffix[i] = suffix[i + nlen - 1 - j];
+       /* Otherwise ... */
+       else {
+          if (i < k)
+             k = i;
+          j = i;
+          /* Compare needle suffix with suffix of needle up to i */
+          while (k >= 0 && needle[k] == needle[k + nlen - 1 - j])
+             --k;
+          /* Set suffix to the number of chars in suffixes comparing equal */
+          suffix[i] = j - k;
+       } 
+       i--;
+    }
+
+    /* 
+     *  Create good suffix table, phase two: Creating the skip table.
+     *  This table tells us how long we can skip before we have a
+     *  reoccurrence of an already matched suffix.
+     */
+
+    /* Initialize skip table to skip a full needle length */
+
+    for (i = 0; i < nlen; ++i)
+       skip[i] = nlen;
+
+    j = 0;
+
+    /* Loop over needle and check for the special case when we have a 
+     * suffix at i reaching the edge of the needle if so, ok to skip
+     * all the way to the edge but no more */
+
+    for (i = nlen - 1; i >= -1; --i)
+       if (i == -1 || suffix[i] == i + 1)
+          for (; j < nlen - 1 - i; ++j)
+             if (skip[j] == nlen)
+                skip[j] = nlen - 1 - i;
+
+    /* Each position is awarded skips according to when a suffix reoccurs.
+     * For each suffix length v in the suffix table (remember, the value v 
+     * at a position i in the suffix table is the length of a reoccuring 
+     * suffix of the needle at i, 0 at i means that no suffix of the needle is 
+     * reoccuring at position i) we want to find the highest i for which this
+     * value v occurs, i.e. we want to find the latest position in the
+     * string where we had the suffix reoccuring. In other words:
+     *
+     * skip[needle_len-1 - v] = 
+     *           needle_len-1 - highest i for which a suffix of length v exists
+     * 
+     */
+
+    if (nlen > 1)
+      for (i = 0; i <= nlen - 2; ++i)
+         skip[nlen - 1 - suffix[i]] = nlen - 1 - i;
+
+    /* Some examples:
+     *
+     * Suffix "bc" is also a suffix at i=4, the length of the common suffix 
+     * is 2 thus suffix[i=4] is 2. If we so far have matched "bc" and do 
+     * not match "a" we shift the needle 
+     * skip[(nlen-1-successfully_matched_chars)=9-1-2=6] = 4 steps forward 
+     * where we have the next occurrence of "bc". 
+     *
+     * If we match "c" and fail to match "b" we get to skip a full 
+     * skip[(nlen-1-successfully_matched_chars)=9-1-1=7] = 9 steps. 
+     * While the "c" suffix exists earlier in the string it is always 
+     * with a "b" before it, and we have already failed to match "bc".
+     *
+     * i      0 1 2 3 4 5 6 7 8
+     * needle x x y b c x a b c
+     * suffix 0 0 0 0 2 0 0 0 9
+     * skip   9 9 9 9 9 9 4 9 1
+     *
+     * Here is the same example where the suffix "c" occurs twice. Observe 
+     * that if we match "c" but fail to match "bc" we get to jump 7 steps 
+     * to the c in position 1. 
+     *
+     * i      0 1 2 3 4 5 6 7 8
+     * needle x c y b c x a b c
+     * suffix 0 1 0 0 2 0 0 0 9
+     * skip   9 9 9 9 9 9 4 7 1
+     *
+     * Here is an example with a suffix continuing over the edge. If we match
+     * "abc" and then fail to match x we cannot jump more than 8 chars since
+     * the c at the beginning of the needle may match.
+     *
+     * i      0 1 2 3 4 5 6 7 8
+     * needle c x y b c x a b c
+     * suffix 1 0 0 0 2 0 0 0 9
+     * skip   8 8 8 8 8 8 4 8 1
+     *
+     */
+
+    /* 
+     *  Now use the preprocessed tables and do the actual search
+     *  Advance step by step while we have a match until end of
+     *  needle. If no match we get to skip forward MAX of the values 
+     *  in the precalculated tables. 
+     */
+
+    for(hpos = 0; hpos < hlen - nlen + 1; ) {
+        int npos = nlen - 1;
+        while(needle[npos] == haystack[npos + hpos]) {
+            if(npos == 0) 
+        	 BIF_RET(make_small(hpos+startpos));
+            --npos;
+        }
+        hpos += MAX(skip[npos], npos - occ[haystack[npos + hpos]]);
+    }
+    BIF_RET(make_small(0));
+}
+
+
+/*
+ *  For multi-pattern search we use a Aho-Corasick style algorithm, a good 
+ *  balance between speed and simplicity... 
+ *
+ *  We build a trie with a node for each letter in needles which are not 
+ *  sharing a common suffix. For search patterns sharing a common suffix
+ *  e.g. abc and abd we will only need 4 nodes, a, b, c and d.
+ */
+
+typedef struct node node_t;
+struct node {
+  node_t *fail;          // Which node to go to on fail
+  node_t *parent;        // My parent
+  int depth;             // The level in the trie
+  unsigned char symbol;  // Char associated with edge to this node
+  int key_id;            // Search pattern ending on this node
+  int key_len;           // Length of search pattern ending on this node
+  node_t *next[256];     // Which node to go to on match for a character
+  int id; //For debugging
+};
+
+/*
+ * We add each keyword in a multi-pattern search to a trie.
+ * If the suffixes are the same as for an already existing keyword
+ * we do not create new nodes for the letters in the already existing
+ * suffix.
+ */
+
+void add_keyword(node_t *root, unsigned char *key, 
+                 int key_len, int key_id, node_t *nodes) {
+  node_t *current_node = root;
+  node_t *new_node = nodes;
+  int j = 0;
+
+//  (printf("\nSearch key %d: \r\n", key_id));                // For debugging
+
+  /*
+   * Add nodes for each letter in keyword which does not already have one
+   * A lot of values are intentionally not initialized. They are set to
+   * zero when calling calloc to initialize the struct
+   */
+
+  for (j = 0; j < key_len ; j++) {
+     if (current_node->next[*(key+j)] == NULL) {
+
+//        new_node->id = key_id*100+j;                        // For debugging
+//        (printf(" [%c, %d] \r\n", *(key+j),new_node->id )); // For debugging
+
+        new_node->symbol = *(key+j);
+        new_node->depth = j;
+        new_node->parent = current_node;
+        current_node->next[*(key+j)] = new_node++;
+     }
+//     else (printf(" %c \r\n", *(key+j)));                   // For debugging
+
+     current_node = current_node->next[*(key+j)];
+  }
+
+  /* Some settings for the final node in a search key*/
+  current_node->key_id = key_id;
+  current_node->key_len = key_len;
+}
+
+/*
+ * Optimize the trie by adding pointers to follow when we fail to match
+ * a character. This needs to be done on a level by level base, i.e.
+ * we cannot take care of the nodes on depth = 2 before we have done the
+ * nodes directly under the root.  
+ */
+
+void set_failpointers(node_t *root, int no_of_nodes)
+{
+   int i, j, changes = 1, current_depth = 1;
+   node_t *node, *tmp_node, *parent;
+   unsigned char symbol;
+
+   /* failure pointers for depth 0 and root points to root */
+     node = root;
+     for(i = 0; i < no_of_nodes; i++, node++) 
+       if ( (node != NULL) && (node->depth == 0) ) {
+           node->fail = root;
+//          (printf("fail(%d)->root\r\n",node->id));               // For debug
+       } 
+
+   while(changes > 0) {
+     changes = 0;
+     node = root;
+     for(j = 1; j < no_of_nodes; j++, node++) 
+       if ( (node != NULL) && (node->depth == current_depth) ) {
+           changes++;
+           parent = node->parent;
+           symbol = node->symbol;
+
+           /* traverse tree upwards by following failure functions until we */ 
+           /* find a node which has transition function for our symbol */ 
+
+           tmp_node = parent->fail;
+           while( (tmp_node != root) && (tmp_node->next[symbol] == NULL) ) 
+              tmp_node = tmp_node->fail;
+          
+           /* if we found a node with transition pointer for our symbol */
+           /* set it, otherwise we jump to root at failure */
+          
+           if ( tmp_node->next[symbol] != NULL )
+                node->fail = tmp_node->next[symbol];
+           else
+                node->fail = root;
+
+//          (printf("fail(%d)->%d\r\n",node->id, node->fail->id)); // For debug
+       }
+     current_depth++;
+   }
+}
+
+
+BIF_RETTYPE multi_pattern_search(Process* A__p, node_t *root, byte* haystack, 
+			 Uint startpos,	 Uint hlen, Uint no_of_nodes) 
+{
+    Uint j;
+    node_t *node = root;
+    node_t *new_node;
+
+    set_failpointers(root, no_of_nodes);
+
+    for(j = 0; j < hlen; j++) {
+
+      /* While no match, follow fail pointer until we have a match (or root) */
+
+      while ( node != root && ( new_node = node->next[haystack[j]] ) == NULL )
+         node = node->fail;
+
+      /* If we have a match, follow transition to the corresponding node */
+
+      if ( node->next[haystack[j]] != NULL ) 
+         node = node->next[haystack[j]];
+
+      /* If the current node is the end of a key, return matching position */
+
+      if ( node->key_id != 0 ) {
+          Eterm* hp;
+	  int ret = j - node->key_len + 2;
+//	  (printf("Match at %d, search key %d\r\n", ret, node->key_id));
+
+          hp = HAlloc(BIF_P, 3);
+          BIF_RET(TUPLE2(hp, make_small(node->key_id),
+                             make_small(ret+startpos-1)));
+      }
+  }
+  return make_small(0);
+}
+
+
+BIF_RETTYPE match_binary_3(Process* p, Eterm Arg1, Eterm Arg2, Eterm Arg3)
+{
+    Eterm real_bin, obj;
+    byte *haystack, *needle;
+    Uint hlen, nlen, bitsize, bitoffs, offset, startpos, endpos;
+    Eterm* tp;
+
+    if (is_not_tuple(Arg3)) goto error;
+
+    tp = tuple_val(Arg3);
+    if (tp[0] != make_arityval(2)) goto error;
+    if (!term_to_Uint(tp[1], &startpos)) goto error;
+    if (!term_to_Uint(tp[2], &endpos)) goto error;
+
+    if (is_not_binary(Arg1)) goto error;
+
+    hlen = binary_size(Arg1);
+    if (hlen == 0) goto no_match;
+
+    if (startpos < 1 || startpos > hlen) goto error;
+    if (endpos < startpos || endpos > hlen) goto error;
+
+    hlen = endpos - startpos + 1;
+
+    ERTS_GET_REAL_BIN(Arg1, real_bin, offset, bitoffs, bitsize);
+    haystack = binary_bytes(real_bin)+offset+startpos-1;
+
+    if (is_list(Arg2)) {
+       int key_id = 1;
+       int max_node = 1; 
+       int total_nlen = 0;
+       node_t *root;
+       BIF_RETTYPE r;
+       Eterm List = Arg2;
+
+       while(is_list(List)) {
+         obj = CAR(list_val(List));
+         if (is_binary(obj)) total_nlen += binary_size(obj);
+         else goto error;         
+         List = CDR(list_val(List));
+       } 
+
+       /* If we have a lot of common suffixes we might not use all of this
+        * heap but it is a pretty good upper limit for how much we will need.
+        * Allocing all mem at once rather than once per node is a *lot* faster
+        *
+        * NB NB NB: calloc clears memory, if replaced with malloc
+        * strange badness will occur due to unitialized variables... (yes,
+        * Captain Obvious is a friend of mine, but you have been warned)*/
+
+       root = calloc(total_nlen+1, sizeof(node_t));
+       assert(root != NULL);
+
+       while(is_list(Arg2)) {
+         obj = CAR(list_val(Arg2));
+
+         nlen = binary_size(obj);
+    
+         if (nlen > 0) {
+            ERTS_GET_REAL_BIN(obj, real_bin, offset, bitoffs, bitsize);
+            needle = binary_bytes(real_bin)+offset;
+            add_keyword(root, needle, nlen, key_id++, &(root[max_node]));
+            max_node += nlen;
+         }
+
+         Arg2 = CDR(list_val(Arg2));
+       }
+       r = multi_pattern_search(p, root, haystack, startpos, hlen, max_node);
+       free(root);
+       BIF_RET(r);
+    }
+    else if (is_binary(Arg2)) {
+        nlen = binary_size(Arg2);
+
+        if (nlen > hlen || nlen == 0)
+           BIF_RET(make_small(0));
+
+        ERTS_GET_REAL_BIN(Arg2, real_bin, offset, bitoffs, bitsize);
+        needle = binary_bytes(real_bin)+offset;
+
+        /* Preprocessing costs. Use brute force if needle*haystack is small. */
+
+        if ( hlen*nlen < 256 ) {
+           int i,j;
+           for (i = 0; i < hlen; i++) {
+              for (j = 0; j < nlen; j++)
+                 if ( (i+j > hlen-1) || (needle[j] != haystack[i+j]) ) break;
+              if (j == nlen) BIF_RET(make_small(startpos+i));
+           }
+           goto no_match;
+        }
+
+        BIF_RET(single_pattern_search(haystack, needle, startpos, hlen, nlen));
+    }
+  error:
+    BIF_ERROR(p, BADARG);
+  no_match:
+    BIF_RET(make_small(0));
+}
+
+
+/*** REGEX starts here ***/
+
+#define RegexOr 300
+#define RegexFinal 400
+#define RegexEOS 500
+#define RegexBOS 600
+#define RegexEpsilon 700
+
+typedef struct regex_node regex_node_t;
+struct regex_node {
+  regex_node_t *next;     // Which node to go to on match for a character
+  regex_node_t *alt_next; // Alternative node used by "|"
+  unsigned int symbol;    // Char associated with edge to this node
+  Uint generation;        // Generation this node was last added to state-list
+  Uint start_pos;         // Start position of search pattern ending on this node
+  Uint next_start_pos;    // Start pos for next gen of this node
+  int id; //For debugging
+};
+
+typedef struct regex_state regex_state_t;
+struct regex_state {
+  regex_node_t *node;     // Node associated with this state
+};
+
+
+regex_node_t *new_node(regex_node_t *nodes, int symbol, 
+                       regex_node_t *next, regex_node_t *alt_next, int *id){
+   regex_node_t * node = ++(*id) + nodes;
+   node->symbol = symbol;
+   node->id = *id;
+   node->next = next;
+   node->generation = UINT_MAX;
+   node->start_pos = UINT_MAX;
+   node->next_start_pos = UINT_MAX;
+   node->alt_next = alt_next;
+   return node;
+}
+
+regex_node_t *decode(regex_node_t *nodes, Eterm Arg, 
+                     regex_node_t *next, int *id) {
+   if (is_tuple(Arg)) {
+      Eterm* tp = tuple_val(Arg);
+      if (is_atom(tp[1])) {
+
+         if ( (tp[1] == am_ok) && (arityval(*tp) == 2) )
+             return decode(nodes, tp[2], next, id);
+
+         /* tp[2] followed by tp[3] */
+         else if ( (tp[1] == am_concat) && (arityval(*tp) == 3) ) {
+             regex_node_t *to = decode(nodes, tp[3], next, id);
+             regex_node_t *from = decode(nodes, tp[2], to, id);
+             return from;
+         }
+
+         /* tp[2] | tp[3] */
+         else if ( (tp[1] == am_or) && (arityval(*tp) == 3) ) 
+             return new_node(nodes, RegexOr, 
+                             decode(nodes, tp[2], next, id),
+                             decode(nodes, tp[3], next, id),
+                             id);
+
+
+         /* ? zero or one */
+         else if ( (tp[1] == am_optional) && (arityval(*tp) == 2) ) 
+             return new_node(nodes, RegexOr, next,
+                             decode(nodes, tp[2], next, id), id);
+
+
+         /* * zero or more, slightly tricky, node refers to itself */
+         else if ( (tp[1] == am_kclosure) && (arityval(*tp) == 2) ) {
+             regex_node_t *node = ++(*id) + nodes;
+             node->symbol = RegexOr; 
+             node->id = *id;  //DEBUG
+             node->generation = UINT_MAX;
+             node->start_pos = UINT_MAX;
+             node->next_start_pos = UINT_MAX;
+             node->next = next;
+             node->alt_next = decode(nodes, tp[2], node, id);
+             return node;
+         }
+
+         /* + one or more, slightly tricky as node refers to itself */
+         else if ( (tp[1] == am_pclosure) && (arityval(*tp) == 2) ) {
+             regex_node_t *from;
+             regex_node_t *node = ++(*id) + nodes;
+             node->symbol = RegexOr; 
+             node->id = *id;  //DEBUG
+             node->generation = UINT_MAX;
+             node->start_pos = UINT_MAX;
+             node->next_start_pos = UINT_MAX;
+             from = decode(nodes, tp[2], node, id);
+             node->next = from;
+             node->alt_next = next;
+             return from;
+         }
+
+
+         // TODO
+         /* [a-z] */
+         else if ( (tp[1] == am_char_class) && (arityval(*tp) == 2) ) {
+         }
+
+         // TODO
+         /* [^a-z], this also covers "." which is comp_class of "\n" */
+         else if ( (tp[1] == am_comp_class) && (arityval(*tp) == 2) ) {
+         }
+
+         /* We have already checked syntax, we should never get here */
+         else assert( 0 == 1 ); 
+     }
+
+   } else if (is_small(Arg)) 
+        return new_node(nodes, unsigned_val(Arg), next, NULL, id);
+     else if (is_atom(Arg) && Arg == am_bos)
+        return new_node(nodes, RegexBOS, next, NULL, id);
+     else if (is_atom(Arg) && Arg == am_eos)
+        return new_node(nodes, RegexEOS, next, NULL, id);
+     else if (is_atom(Arg) && Arg == am_epsilon)
+        return new_node(nodes, RegexEpsilon, next, NULL, id);
+
+   /* We have already checked syntax, we should never get here */
+   assert( 0 == 1 );
+   return NULL; 
+}
+
+
+int check_syntax(Eterm Arg) {
+   if (is_tuple(Arg)) {
+      Eterm* tp = tuple_val(Arg);
+      if (is_atom(tp[1])) {
+
+         if ( (tp[1] == am_ok) && (arityval(*tp) == 2) )
+             return check_syntax(tp[2]);
+
+         /* tp[2] followed by tp[3] */
+         else if ( (tp[1] == am_concat) && (arityval(*tp) == 3) ) 
+             return check_syntax(tp[3]) + check_syntax(tp[2]);
+
+         /* tp[2] | tp[3] */
+         else if ( (tp[1] == am_or) && (arityval(*tp) == 3) ) 
+             return 1 + check_syntax(tp[2]) + check_syntax(tp[3]);
+
+         /* ? zero or one */
+         else if ( (tp[1] == am_optional) && (arityval(*tp) == 2) ) 
+             return 1 + check_syntax(tp[2]);
+
+         /* * zero or more, slightly tricky, node refers to itself */
+         else if ( (tp[1] == am_kclosure) && (arityval(*tp) == 2) ) 
+             return 1 + check_syntax(tp[2]);
+
+         /* + one or more, slightly tricky, node refers to itself */
+         else if ( (tp[1] == am_pclosure) && (arityval(*tp) == 2) ) 
+             return 1 + check_syntax(tp[2]);
+
+         /* [a-z] */
+         // TODO
+         else if ( (tp[1] == am_char_class) && (arityval(*tp) == 2) ) {
+         }
+
+         /* [^a-z], this also covers "." which is comp_class of "\n" */
+         // TODO
+         else if ( (tp[1] == am_comp_class) && (arityval(*tp) == 2) ) {
+         }
+
+         return INT_MIN; 
+     }
+
+   } 
+   else if (is_small(Arg)) return 1;
+   else if (is_atom(Arg) && Arg == am_bos) return 1;
+   else if (is_atom(Arg) && Arg == am_eos) return 1;
+   else if (is_atom(Arg) && Arg == am_epsilon) return 1;
+
+   return INT_MIN; 
+}
+
+/* Check for a possible match and add if start position is
+ * earlier than a previously found possible match */
+
+int is_final(regex_node_t *node, Uint i, Uint *p_match, 
+             Uint *p_len, Uint start_pos, Uint len){
+
+   if ((node->symbol == RegexFinal) && ( start_pos <= *p_match)) {
+       *p_match = start_pos;
+       if (*p_len < i - start_pos + len) 
+           *p_len = i - start_pos + len;
+       return 1;
+   }
+   return 0;
+}
+
+/* Add a node to the list of active states */
+ 
+Uint add(regex_state_t *state, regex_node_t *node, Uint start_pos, 
+         Uint index, Uint generation) {
+
+   /* If this node has already been added this generation, don't add
+    * it again. If we do we will have exponential behaviours for
+    * certain regexp:s. If it already exists we modify it instead */
+   if (node->generation == generation && generation != UINT_MAX) {
+       /* Don't modify start_pos if the new one isn't better than the old */
+       if (start_pos < node->next_start_pos)
+           node->next_start_pos = start_pos;
+       node->generation = generation;
+       state->node = node;
+       return index;
+   } else {
+       node->next_start_pos = start_pos;
+       node->generation = generation;
+       state->node = node;
+       return ++index;
+   }
+}
+
+/* Recursively check all states for matches */
+
+void check_state(Uint i, byte *haystack, Uint *p_match, Uint *p_len, 
+                 Uint start_pos, regex_state_t **n_states, Uint *n_len, 
+                 regex_node_t *node) {
+
+    Uint symbol = node->symbol;
+    if (start_pos == UINT_MAX) start_pos = i;
+
+    if (symbol == RegexOr) {
+       if (!is_final(node->next, i, p_match, p_len, start_pos,0))
+         check_state(i, haystack, p_match, p_len, start_pos,
+                     n_states, n_len, node->next);
+
+       if (!is_final(node->alt_next, i, p_match, p_len, start_pos,0))
+         check_state(i, haystack, p_match, p_len, start_pos, 
+                     n_states, n_len, node->alt_next);
+    }
+
+    else if (symbol == haystack[i]) {
+      if (!is_final(node->next, i, p_match, p_len, start_pos, 1))
+         *n_len = add(*n_states+*n_len, node->next, start_pos, *n_len, i);
+    }
+}
+
+/* Check any remaining states after we have run the NFA 
+ * We might have a better match at the end */
+
+void check_final(Uint i, Uint *p_match, Uint *p_len, 
+                 Uint start_pos, regex_node_t *node) {
+
+    is_final(node, i, p_match, p_len, start_pos, 1);
+
+    if (node->symbol == RegexOr) {
+       if (!is_final(node->next, i, p_match, p_len, start_pos,0))
+          check_final(i, p_match, p_len, start_pos, node->next);
+
+       if (!is_final(node->alt_next, i, p_match, p_len, start_pos,0))
+          check_final(i, p_match, p_len, start_pos, node->alt_next);
+    }
+}
+
+BIF_RETTYPE regex_binary_3(Process* p, Eterm Arg1, Eterm Arg2, Eterm Arg3)
+{
+    Eterm real_bin;
+    byte *haystack;
+    regex_node_t *root, *nodes; 
+    regex_state_t *temp=NULL, *c_states=NULL, *n_states=NULL;
+    Uint hlen, bitsize, bitoffs, offset, startpos, endpos; 
+    Uint i, j;
+    Uint p_match=UINT_MAX, p_len=0; // Best possible match, len found so far
+    Uint c_len, n_len;              // Length of state arrays
+    Uint alloc_states;
+    int no_of_nodes1=0, no_of_nodes2=0;
+    Eterm* hp;
+
+    Eterm* tp;
+    if (is_not_tuple(Arg3)) goto badarg;
+
+    tp = tuple_val(Arg3);
+    if (tp[0] != make_arityval(2)) goto badarg;
+    if (!term_to_Uint(tp[1], &startpos)) goto badarg;
+    if (!term_to_Uint(tp[2], &endpos)) goto badarg;
+
+    if (is_not_tuple(Arg2)) goto badarg;
+    if (is_not_binary(Arg1)) goto badarg;
+
+    hlen = binary_size(Arg1);
+    if (hlen == 0) goto no_match; //This is compatible with regexp:match/2
+
+    if (startpos < 1 || startpos > hlen) goto badarg;
+    if (endpos < startpos || endpos > hlen) goto badarg;
+
+    hlen = endpos - startpos + 1;
+
+    ERTS_GET_REAL_BIN(Arg1, real_bin, offset, bitoffs, bitsize);
+    haystack = binary_bytes(real_bin)+offset+startpos-1;
+
+    no_of_nodes1 = check_syntax(Arg2);
+    if ( no_of_nodes1 < 0 ) goto badarg;
+    nodes = calloc(no_of_nodes1+1, sizeof(regex_node_t));    
+ 
+    /* Let nodes[0] be the final node */
+    nodes->symbol = RegexFinal;
+    nodes->id = 0;    //DEBUG
+    nodes->next = NULL;
+
+    /* Now build the tree representing the NFA that we are going to run */
+    root = decode(nodes, Arg2, &(nodes[0]), &no_of_nodes2);
+    if (root == NULL) {free(nodes); goto badarg;}
+
+    if (root->symbol == RegexEpsilon) goto epsilon;
+
+    /* Try to make sure check_syntax(...) is updated whenever decode(...) is */
+    assert(no_of_nodes1 == no_of_nodes2);
+
+    /* Allocate two structures for the states that we are in now 
+     * and the state we will be in once we step forward one step
+     * We use two arrays and swap them so we do not have to allocate
+     * a new array every iteration */
+
+    alloc_states = no_of_nodes1+1;
+    c_states = malloc(alloc_states*sizeof(regex_state_t));
+    n_states = malloc(alloc_states*sizeof(regex_state_t));
+
+    /* Initially only the root=start node is part of the states tables */
+    c_len = add(c_states, root, UINT_MAX, 0, UINT_MAX);
+    
+    //TODO: Handle RegexBOS and RegexEOS, slightly unclear how. Do we want
+    //to be compatible with strings and let (<<"a\nb">>, "a$") match?
+    //The regexp module is inconsistent, why does not ("b\na", "^a") match if
+    //("a\nb", "a$") matches?
+
+    for(i=0; i < hlen; i++) {
+      n_len = 0; 
+
+      /* Check all active states in the NFA */
+      for(j=0; j < c_len; j++ ){
+         regex_node_t *node = (c_states+j)->node;
+         Uint start_pos = node->start_pos;
+
+         check_state(i, haystack, &p_match, &p_len, start_pos, 
+                     &n_states, &n_len, node);
+      }
+
+      /* Swap the arrays for next and current state */
+      temp = c_states;
+      c_states = n_states;
+      n_states = temp;
+      c_len = n_len;
+
+      if (n_len == 0 && p_match != UINT_MAX) goto check_final; 
+      
+      /* Add the first state if not already added */
+      c_len = add(c_states+c_len, root, UINT_MAX, c_len, i);
+
+      /* Make sure start_pos for all active nodes is updated simultaneously */
+      for (j=0; j < c_len; j++) 
+         (c_states+j)->node->start_pos = (c_states+j)->node->next_start_pos;
+    }
+
+ check_final:
+    /* Check remaining states, if any */
+    if (c_len > 0) 
+      for( j=0; j < c_len; j++ ) {
+         regex_node_t *node = (c_states+j)->node;
+         Uint start_pos = node->start_pos;
+
+         check_final(i, &p_match, &p_len, start_pos, node);
+      }
+
+    if (p_match != UINT_MAX) goto match; 
+
+    free(c_states); free(n_states); free(nodes);
+
+ no_match:
+    /* Only goto here before alloc */
+    BIF_RET(make_small(0));
+
+ badarg:
+    /* Only goto here before alloc */
+    BIF_ERROR(p, BADARG);
+
+ /* The regexp module returns {match,1,0} on epsilon (unless the haystack 
+  *  is empty as well) */ 
+ epsilon:
+    free(nodes);
+    hp = HAlloc(p, 3);
+    BIF_RET(TUPLE2(hp, make_small(1),make_small(0)));
+
+ match:
+    free(c_states); free(n_states); free(nodes);
+    hp = HAlloc(p, 3);
+    BIF_RET(TUPLE2(hp, make_small(p_match+1),make_small(p_len)));
+
+}
+
